"Main entrypoint logic for cpplint hook."
import argparse
import logging
import os
import subprocess
import sys
import typing

import precommit_diffcheck
import unidiff

import cpplinthook.cpplint
import cpplinthook.formatters

def main() -> None:
	"Main entrypoint for the hook."
	parser = argparse.ArgumentParser()
	parser.add_argument("--fail",
		action="store_true", help="Force failure, useful to see logging.")
	parser.add_argument("--root",
		help="The root directory to use when calculating header guard names.")
	parser.add_argument("--header-guard-base",
		help="An additional base to add to header guards.")
	parser.add_argument("--verbose", action="store_true", help="Show verbose logging.")
	parser.add_argument("filenames", nargs="*", help="The file(s) to lint.")
	args = parser.parse_args()

	logging.basicConfig(level=logging.DEBUG if args.verbose else logging.WARNING)

	_configure_cpplint(args.root, args.header_guard_base)
	has_errors = _process(args.filenames)
	sys.exit(1 if has_errors or args.fail else 0)


def _configure_cpplint(root: str, base: str) -> None:
	"""Set up configuration for cpplint.

	Args:
		root: The root of the code repository for the purpose of calculating header guards.
	"""
	cpplinthook.cpplint.state().ResetErrorCounts() # type: ignore
	if base:
		cpplinthook.cpplint.state().SetHeaderGuardBase(base) # type: ignore
	if root:
		root = os.path.abspath(root)
		cpplinthook.cpplint.state().SetRoot(root) # type: ignore


def _process(filenames: typing.Iterable[str]) -> bool:
	"""Process the files in the current changeset.

	This will print information on the cpplint violations of the current
	changeset. That is, either the currently staged files or the last
	commit.

	Returns true if there are lint violations.

	Args:
		filenames: The filenames to consider, or all of them if an empty list.
	"""
	patchset = precommit_diffcheck.get_diff_or_content(filenames)

	error_count = 0
	for patch in patchset:
		output_formatter = cpplinthook.formatters.DiffFormatter(patch)
		# Remove 'b/' from git patch format
		if patch.target_file.startswith("b/"):
			target = patch.target_file[2:]
		elif patch.target_file == "/dev/null":
			continue
		cpplinthook.cpplint.ProcessFile(target, 0, output_formatter) # type: ignore
		error_count += output_formatter.error_count

	print("Total errors found: {}".format(error_count))
	return error_count > 0



"Custom formatter used to capture detail about failures."
import logging
import typing

import unidiff

import cpplinthook.cpplint

LOGGER = logging.getLogger(__name__)


class FileFormatter(cpplinthook.cpplint.OutputFormatter): # type: ignore
	"""Basic formatter for showing all failures in a file.

	Since we don't control the names from cpplint we just disable some warnings
	from pylint on the various weird names, parameters, etc.
	"""
	def __init__(self):
		self.error_count = 0
		self.has_shown_header = False

	def _Format(self, # pylint: disable=invalid-name, too-many-arguments
		filename: typing.Text,
		linenum: int,
		category:typing.Text,
		confidence: int,
		message: typing.Text):
		"Format a single error in a file."
		if not self.has_shown_header:
			print("{}:".format(filename))
			self.has_shown_header = True
		print("\t{filename}:{linenum}:  {message}  [{category}] [{confidence}]".format(
			category=category,
			confidence=confidence,
			filename=filename,
			linenum=linenum,
			message=message,
		))
		self.error_count += 1

	def FinishedFile(self, filename): # pylint: disable=invalid-name
		"Called when a file is entirely finished."

	def Finished(self): # pylint: disable=invalid-name, no-self-use
		"Called when finished. Not sure how it differs from FinishedFile."
		for category, count in \
			cpplinthook.cpplint.state().errors_by_category.items():
			print("Category '{}' errors found: {}".format(category, count))

class DiffFormatter(FileFormatter):
	"""A formatter for showing errors against a diff."""
	def __init__(self, patch: unidiff.PatchedFile):
		super().__init__()
		self.filename = patch.target_file[2:]
		self.has_failure_outside_diff = False
		self.added_lines: typing.Set[int] = set()
		for hunk in patch:
			# Include only added lines in our analysis
			for i, line in enumerate(hunk.target_lines()):
				if line.is_context:
					continue
				if line.is_added:
					self.added_lines.add(hunk.target_start + i)
		LOGGER.debug(
			"Only accepting failures in %s on lines: %s",
			self.filename, sorted(list(self.added_lines)))


	def _Format(self, # pylint: disable=invalid-name, too-many-arguments
		filename: typing.Text,
		linenum: int,
		category: typing.Text,
		confidence: int,
		message: typing.Text):
		"""Log message is they match against the patch."""
		assert filename == self.filename, "{} is not {}".format(filename, self.filename)
		# Always allow failures on some magical lines since these are hard-coded into cpplint
		if linenum in self.added_lines or linenum == 0:
			super()._Format(filename, linenum, category, confidence, message)
			self.error_count += 1
		else:
			self.has_failure_outside_diff = True
			LOGGER.debug("Found failure outside of diff on line %d: %s", linenum, message)

	def FinishedFile(self, filename):
		"""Message to show when a file is done being processed."""
		if self.has_failure_outside_diff:
			print("\tFailures outside your changes are present in this file. "
				"Run 'pre-commit run cpplint --files {}' to see them".format(filename))



"Logic to set up the module."
import setuptools # type: ignore

setuptools.setup(
    install_requires=[
		"precommit-diffcheck==1.1",
		"six==1.12.0",
		"unidiff==0.5.5",
	],
	extras_require={
		"develop": [
			"mypy",
			"nose2",
			"pylint",
		]
	},
)
